# mad/db.py
#
#

""" JSON file db. """

from bot.event import Event
from bot.get import root
from bot.join import j
from bot.object import Default, Object
from bot.time import elapsed, fn_time, to_day, day
from bot.trace import get_exception

import bot.selector
import logging
import time
import os

class Db(Object):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        
    def scan(self, path, *args, **kwargs):
        """ scan all files. """
        p = Default(kwargs, default="")
        if not path.endswith(os.sep):
            path += os.sep
        result = []
        for root, dirs, files in os.walk(path, topdown=True):
            if not os.path.isdir(root):
                continue
            for fn in files:
                fnn = j(root, fn)
                timed = fn_time(fnn)
                if timed and p.start and timed < p.start:
                    continue
                if timed and p.end and timed > p.end:
                    continue
                yield fnn

    def find(self, prefix, *args, **kwargs):
        """ find all objects stored with a prefix subdirectory. """
        for fn in self.prefixed(prefix, *args):
            try:
                object = Object().load(fn)
            except Exception as ex:
                logging.warn("fail %s %s" % (fn, str(ex)))
                continue
            if "deleted" in object and object.deleted:
                continue
            yield object

    def prefixed(self, *args, **kwargs):
        """ return all filename in a workdir subdirectory, the 'prefix'. """
        from bot.space import cfg
        if not args:
            return []
        files = []
        if cfg.workdirs:
            workdirs = cfg.workdirs
        else:
            workdirs = [root()]
        for workdir in workdirs:
            path = j(workdir, args[0])
            if not os.path.exists(path):
                continue
            for fn in self.scan(path, *args, **kwargs):
                files.append(fn)
        return sorted(files, key=lambda x: fn_time(x))

    def prefixes(self):
        for p in os.listdir(root()):
            yield p

    def is_prefix(self, prefix):
        for p in os.listdir(root()):
            if prefix in p:
                return True

    def selected(self, event):
        """ select objects based on a _parsed event. """
        nr = 0
        thrs = []
        if not event._parsed.args:
            return []
        starttime = time.time()
        for fn in self.prefixed(*event._parsed.args, **event._parsed):
            object = self.selector(event, fn)
            if object:
                nr += 1
                if event._parsed.index and event._parsed.index != nr:
                    continue
                yield object
        endtime = time.time()
        logging.warn("# selected %s %s" % (nr, elapsed(endtime - starttime, short=False)))
       
    def selector(self, event, fn, object=None):
        s = bot.selector
        if not object:
            try:
                object = Object().load(fn)
            except Exception as ex:
                logging.warn("fail %s %s" % (fn, get_exception()))
                return
        if "nodel" not in event and "deleted" in object and object.deleted:
            return
        if not s.selector(object, event._parsed.fields):
            return
        if s.notwanted(object, event._parsed.notwant):
            return
        if not s.wanted(object, event._parsed.want):
            return
        if s.ignore(object, event._parsed.ignore):
            return
        if s.uniq(object, event._parsed.uniq):
            return
        return object

    def sequence(self, prefix, start, end=time.time(), skip=[]):
        """ select objects of type prefix, start time till end time. """
        p = Object()
        p.start = start
        p.end = end
        for fn in self.prefixed(prefix, **p):
            do_skip = False
            for k in skip:
                if k in fn:
                    do_skip = True
            if do_skip:
                continue
            try:
                e = Event().load(fn)
            except Exception as ex:
                logging.warn("fail %s %s" % (fn, str(ex)))
                continue
            yield e

    def since(self, start, *args, **kwargs):
        """ return all objects since a given time. """
        e = Event(**kwargs)
        e.start = start
        for fn in self.prefixed(*args, **e):
            try:
                object = Object().load(fn)
            except Exception as ex:
                logging.warn("fail %s %s" % (fn, str(ex)))
                continue
            if "deleted" in object and object.deleted:
                continue
            yield object
 
    def first(self, *args, **kwargs):
        """ return first object matching provided prefix. """
        for fn in self.prefixed(*args, **kwargs):
            try:
                object = Object().load(fn)
            except Exception as ex:
                logging.warn("fail %s %s" % (fn, str(ex)))
                continue
            if "deleted" in object and object.deleted:
                continue
            if len(args) > 1 and object.get(args[0]) != args[1]:
                continue
            return object

    def last(self, *args, **kwargs):
        """ return last record with a matching prefix. """
        prefix = args[0]
        if len(args) > 1:
            value = args[1]
        else:
            value = ""
        for fn in self.prefixed(args[0], **kwargs)[::-1]:
            try:
                object = Object().load(fn)
            except Exception as ex:
                logging.warn("fail %s %s" % (fn, str(ex)))
                continue
            if "deleted" in object and object.deleted:
                continue
            if value and object.get(prefix, "") != value:
                continue
            return object

def rm(event):
    """ set deleted flag on objects. """
    db = Db()
    nr = 0
    for object in db.selected(event):
        object.deleted = True
        object.sync()
        nr += 1
    event.ok(nr)

def restore(event):
    """ set deleted=False in selected records. """
    db = Db()
    nr = 0
    event.nodel = True
    for object in db.selected(event):
        object.deleted = False
        object.sync()
        nr += 1
    if not nr:
        event.reply("no results")
    event.ok(nr)

def week(event):
    """ show last week's logged objects. """
    db = Db()
    event._parsed.start = to_day(day()) - 7 * 24 * 60 * 60
    event._parsed.end = time.time()
    nr = 0
    for object in db.selected(event):
        nr += 1
        event.display(object, event._parsed.args, str(nr))
    if not nr:
        event.reply("no results")

def today(event):
    """ show objects logged for today.  """
    db = Db()
    event._parsed.start = to_day(day())
    event._parsed.end = time.time()
    nr = 0
    for object in db.selected(event):
        nr += 1
        event.display(object, event._parsed.args, str(nr))
    if not nr:
        event.reply("no results")

def yesterday(event):
    """ show objects added yesterday. """
    db = Db()
    event._parsed.start = to_day(day()) - 24 * 60 * 60
    event._parsed.end = to_day(day())
    nr = 0
    for object in db.selected(event):
        nr += 1
        event.display(object, event._parsed.args, str(nr))
    if not nr:
        event.reply("no results")

def sequence(event):
    """ show sequence of events. """
    db = Db()
    if "end" not in event._parsed:
        end = to_day(day())
    else:
        end = event._parsed.end
    if "start" not in event._parsed:
        start = 0
    else:
        start = event._parsed.start
    nr = 0
    for o in db.sequence(start, end, skip=["feeds", "seen"]):
        nr += 1
        event.reply(o.nice())
    if not nr:
        event.reply("no results")

def find(event):
    """ present a list of objects based on prompt input. """
    nr = 0
    db = Db()
    for object in db.selected(event):
        nr += 1
        event.display(object, event._parsed.args, str(nr), direct=True)
    if not nr:
        event.reply("no results")

def deleted(event):
    """ show deleted records. """
    event.nodel = True
    nr = 0
    db = Db()
    for object in db.selected(event):
        nr += 1
        event.display(object, event._parsed.args, str(nr))
    if not nr:
        event.reply("no results")

def dump(event):
    """ dump objects matching the given criteria. """
    nr = 0
    db = Db()
    for object in db.selected(event):
        nr += 1
        event.reply(object.nice())
    if not nr:
        event.reply("no results")

def first(event):
    """ show the first record matching the given criteria. """
    db = Db()
    object = db.first(*event._parsed.args)
    if object:
        event.reply(object.nice())
    else:
        event.reply("no result")

def last(event):
    """ show last objectect matching the criteria. """
    if not event._parsed.args:
        event.reply("last <prefix> <value>")
        return
    db = Db()
    object = db.last(*event._parsed.args)
    if object:
        event.reply(object.nice())
    else:
        event.reply("no result")
