# mad/object.py
#
#

""" JSON file backed objectect with dotted access.  """

from bot.file import cdir
from bot.get import urled, slice, get_path
from bot.join import j
from bot.json import dumped
from bot.misc import locked, headertxt
from bot.name import sname
from bot.signature import verify_signature, make_signature
from bot.time import rtime
from bot.trace import get_exception

import bot.json

import threading
import _thread
import logging
import atexit
import types
import fcntl
import json
import tty
import sys
import os


class Object(dict):

    """
        Dict with dotted access instead of brackets, with json files to sync and load from.

    """

    def __init__(self, *args, **kwargs):
        """ Construct an Object, dotted dictionairy access instead of brackets. """
        super().__init__(*args, **kwargs)

    def __getattribute__(self, name):
        """
            Get attribute and if fail check item access.

        """
        if name == "_path":
            from bot.space import cfg
            p = j(cfg.workdir, sname(self).lower())
            return p
        if name == "url":
            return urled(self)
        if name == "type":
            return sname(self).lower()
        if name == "_type":
            return str(type(self))
        try:
            return super().__getattribute__(name)
        except AttributeError:
            try:
                return self[name]
            except KeyError:
                raise AttributeError(name)

    def __getattr__(self, name):
        """
            | Get missing attribute by name. initialize into an Object if if name is missing from dictionary.
            | Predefined names are _ready and counter. _ready is for waiting on results and counter is for integer simulation. 

        """ 
        if name == "_connected":
            self._connected = Object()
        if name == "_container":
            self._container = Default()
        if name == "_counter":
            self._counter = Default(default=0)
        if name == "_error":
            self._error = Default()
        if name == "_funcs":
            self._funcs = []
        if name == "_ready":
            self._ready = threading.Event()
        if name == "_state":
            self._state = Object()
        if name == "_thrs":
            self._thrs = []
        if name == "_time":
            self._time = Default(default=0.0)
        if name == "cfg":
            self.cfg = Config().template(self.type)
        if name == "counter":
            self.counter = 0
        try:
            return self[name]
        except KeyError:
            raise AttributeError(name)

    def __iadd__(self, y):
        """ Use the counter attribute to increase internal counter. """
        self.counter += y
        return self.counter

    def __isub__(self, y):
        """ Subract on counter. """
        self.counter -= y
        return self.counter

    def __imul__(self, y):
        """ Multiply on counter. """
        self.counter *= y
        return self.counter

    def __repr__(self):
        """ Return module.class as a class type. """
        return '<%s.%s at %s>' % (
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self))
        )

    def __setattr__(self, name, value):
        """ Implement dotted dict access. """
        return self.__setitem__(name, value)

    def __str__(self):
        """ Return a prettified json string. """            
        return self.nice()

    def grep(self, val):
        """ Grep for a matching stringified value, return a Object with those matching values. """
        o = Object()
        for key, value in self.items():
            if val in str(value) and value not in o:
                o[key] = value
        return o

    def id(self, *args, **kwargs):
        return sname(self).lower()

    def json(self, *args, **kwargs):
        """ JSON string dump. """
        return bot.json.dumps(self, *args, **kwargs)

    def load(self, path, force=False, skip=[], full=True):
        """ Load a json file into this object. use skip as a list of keys to skip. """
        path = os.path.abspath(path)
        logging.debug("# load %s" % path)
        try:
            ondisk = self.read(path)
            self._container.path = path
        except:
            logging.error(get_exception())
            return self        
        fromdisk = bot.json.loads(ondisk)
        if "signature" in fromdisk:
            if not verify_signature(fromdisk["data"], fromdisk["signature"]) and not force:
                raise bot.error.ESIGNATURE(path)
        if "data" in fromdisk:
            self.update(slice(fromdisk["data"], skip=skip, full=full))
            self._container.update(slice(fromdisk, skip=["data"]))
        return self

    def last(self, *args, **kwargs):
        from bot.space import kernel
        d = kernel.last(sname(self).lower())
        if d:
            self.update(d)
        return self

    def loads(self, s):
        """ Update with deconstructed (dict) json string. """
        self.update(bot.json.loads(s))

    def merge(self, object):
        for k, v in object.items():
            if k in self and not self[k]:
                self[k] = v

    def nice(self, *args, **kwargs):
        """ Return a nicyfied, indent=4, sort_key is True, json dump. """
        return bot.json.dumps(self, indent=4, sort_keys=True)

    def pure(self, *args, **kwargs):
        """ Return a sliced (no _ keys), indent=4, sort_key is True, json dump. """
        return bot.json.dumps(slice(self, full=False), indent=4, sort_keys=True)

    def prepare(self):
        """
            | Prepare the object and return a string containing the "data" part.
            | Keyword can be "prefix" when using a subdirectory.
            | Use "saved" when savestamp need to be different from the "now" timestamp.

        """
        import time
        todisk = Object()
        todisk.data = dumped(slice(self, skip=["_container", "_parsed"], full=True))
        todisk.data._type = str(type(self))
        if "prefix" not in todisk:
            todisk.prefix = sname(self).lower()
        if "saved" not in todisk:
            todisk.saved = time.ctime(time.time())
        try:
            todisk.signature = make_signature(todisk["data"])
        except:
            pass
        try:
            result = bot.json.dumps(todisk, indent=4, ensure_ascii=False, sort_keys=True)
        except TypeError:
            raise bot.error.ENOJSON()
        return result

    def printable(self, keys=[], skip=[], nokeys=False):
        """ Determine from provided keys list and/or from skipping from a skiplist a displayable string from those attributes. """
        keys = keys or self.keys()
        result = []
        for key in keys:
             if key == "default":
                 continue
             if key.startswith("_"):
                 continue
             if key in skip:
                 continue
             if nokeys:
                 result.append("%4s" % self[key])
             else:
                 result.append("%-8s %3s" % (key, self[key]))
        return " ".join(reversed(sorted(result)))

    def read(self, path):
        """ Read a json dump from given path, returning the json string with comments stripped. """
        try:
            f = open(path, "r", encoding="utf-8")
        except FileNotFoundError:
            return "{}"
        res = ""
        for line in f:
            if not line.strip().startswith("#"):
                res += line
        if not res.strip():
            return "{}"
        f.close()
        return res

    def register(self, key, val, force=False):
        """ Register key, value and throw an exception is value is already set. """
        if key in self and not force:
            raise bot.error.ESET(key)
        self[key] = val   

    def save(self, stime=""):
        """ save a static (fix filepath) version of this object. """
        return self.sync(stime=stime)

    def search(self, name):
        """ Search this objects keys skipping keys that start with "_". """
        o = Object()
        for key, value in self.items():
            if key.startswith("_"):
                continue
            if key in name:
                o[key] = value
            elif name in key.split(".")[-1]:
                o[key] = value
        return o

    @locked
    def sync(self, path="", stime=""):
        """ Sync to disk using provided/created path. Optionally a path can be provided. """
        from bot.space import cfg, kernel, EBORDER
        if not path:
            path = get_path(self)
        if not path:
            path = j(self._path, stime or rtime())
        path = os.path.abspath(os.path.normpath(path))
        logging.info("! sync %s" % path)
        if cfg.workdir not in path:
            raise EBORDER(path)
        self._container.path = path
        d, fn = os.path.split(path)
        cdir(d)
        todisk = self.prepare()
        rpath = path + "_tmp"
        datafile = open(rpath, 'w')
        fcntl.flock(datafile, fcntl.LOCK_EX | fcntl.LOCK_NB)
        datafile.write(headertxt % "%s characters" % len(todisk))
        datafile.write(todisk)
        datafile.write("\n")
        datafile.flush()
        os.rename(rpath, path)
        fcntl.flock(datafile, fcntl.LOCK_UN)
        datafile.close()
        return path

    def clear(self):
        """ Clear the ready flag. """
        self._ready.clear()

    def isSet(self):
        """ Check whether ready flag is set. """
        return self._ready.isSet()

    def ready(self):
        """ Signal this object as "ready". """
        self._ready.set()

    def wait(self, sec=None):
        """ Wait for this object's ready flag. """
        self._ready.wait(sec)
        for thr in self._thrs:
            thr.join()

class Default(Object):

    """
        A object with a "default" set. Standard default return is Object().

    """

    def __init__(self, *args, **kwargs):
        """ Constructor that initializes a variable with Object() as a default. """
        super().__init__(*args, **kwargs)
        self.default = ""
        self.update(kwargs)

    def __getattr__(self, name):
        """ Override Object.__getattr__. """
        try:
            return super().__getattr__(name)
        except AttributeError:
            self[name] = self.default
        return self[name]

class Config(Default):

    """
        A Config objectect can read previous cfg from disk.

    """

    def template(self, name):
        from bot.template import template
        self.update(template.get(name, {}))
        return self

    def fromdisk(self, name):
        from bot.space import cfg
        self.template(name)
        self.load(j(cfg.workdir, "config", name))
        return self
